package control

import (
	"fmt"
	"io"
	"strings"

	"gopkg.in/yaml.v2"
)

type OutputFormatter interface {
	OpenSection(title string) error
	CloseSection() error

	OpenTable(columns ...string) error
	CloseTable() error

	AddValue(interface{}) error
}

type TextFormatter struct {
	w io.Writer

	chapterDepth int
	inTable      bool
	tableColumns []string
}

func (f *TextFormatter) OpenSection(title string) error {
	f.chapterDepth++
	if _, err := f.w.Write([]byte(
		fmt.Sprintf("%s %s\n\n", strings.Repeat("#", f.chapterDepth), title),
	)); err != nil {
		return err
	}
	return nil
}

func (f *TextFormatter) CloseSection() error {
	f.chapterDepth--
	if _, err := f.w.Write([]byte("\n\n")); err != nil {
		return err
	}
	return nil
}

func (f *TextFormatter) OpenTable(columns ...string) error {
	f.inTable = true
	f.tableColumns = columns
	return nil
}

func (f *TextFormatter) CloseTable() error {
	if _, err := f.w.Write([]byte("\n")); err != nil {
		return err
	}
	f.inTable = false
	return nil
}

func (f *TextFormatter) AddValue(value interface{}) error {
	if f.inTable {
		var sValue string
		if s, ok := value.(string); ok {
			sValue = s
		} else {
			if values, ok := value.([]interface{}); ok {
				var m MapSlice
				for i, v := range values {
					m = append(m, MapItem{
						Key:   f.tableColumns[i],
						Value: v,
					})
				}
				value = m
			}
			b, err := yaml.Marshal(value)
			if err != nil {
				return err
			}
			sValue = string(b)
		}
		if _, err := f.w.Write([]byte(
			fmt.Sprintf("* %s\n", strings.ReplaceAll(sValue, "\n", "\n  ")),
		)); err != nil {
			return err
		}
	} else {
		var output []byte
		if s, ok := value.(string); ok {
			output = []byte(s)
		} else {
			b, err := yaml.Marshal(value)
			if err != nil {
				return err
			}
			output = b
		}
		if _, err := f.w.Write(output); err != nil {
			return err
		}
	}

	return nil
}

type MapItem struct {
	Key   string
	Value interface{}
}

type MapSlice []MapItem

func (s MapSlice) MarshalYAML() (interface{}, error) {
	var ys = make(yaml.MapSlice, len(s))
	for i := range s {
		ys[i].Key = s[i].Key
		ys[i].Value = s[i].Value
	}
	return ys, nil
}
