package director

import (
	"context"
	"fmt"
	"math"
	"strconv"
	"strings"
	"sync"
	"time"

	"github.com/nats-io/nats.go"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/process"
	svc_tools "xbus.io/xbus/services/tools"
	"xbus.io/xbus/storage"
)

const (
	ConfigEntryDelayMax            = "xbus.director.delay.max"
	ConfigEntryDelayIncreaseFactor = "xbus.director.delay.increase-factor"
)

var (
	ConfigEntryDelayMaxDefaultValue            = 30 * time.Minute
	ConfigEntryDelayIncreaseFactorDefaultValue = 100 // 100%
)

// ActorHandlerConfig holds the settings for instanciating a ActorHandler
type ActorHandlerConfig struct {
	Log             zerolog.Logger
	Conn            *nats.Conn
	ActorStorage    storage.ActorStorage
	JobStorage      storage.JobStorage
	ProcessManager  *process.Manager
	EnvelopeStorage storage.EnvelopeStorage
	ActorMap        *svc_tools.ActorMap
	Director        directorServer
	EnvelopeWatcher *EnvelopeWatcher
}

// newActorHandler instanciates a ActorHandler
func newActorHandler(ctx context.Context, actor api.Actor, config ActorHandlerConfig) *ActorHandler {
	delayMax := ConfigEntryDelayMaxDefaultValue
	if s := actor.GetConfigEntry(ConfigEntryDelayMax); s != "" {
		d, err := time.ParseDuration(s)
		if err != nil {
			config.Log.Warn().Err(err).
				Str("actor", actor.Name).
				Str("config-entry", ConfigEntryDelayMax).
				Dur("default-value", ConfigEntryDelayMaxDefaultValue).
				Msg("Could not parse duration. The default value will be used")
		} else {
			delayMax = d
		}
	}

	delayIncreaseFactor := ConfigEntryDelayIncreaseFactorDefaultValue
	if s := strings.TrimSpace(actor.GetConfigEntry(ConfigEntryDelayIncreaseFactor)); s != "" {
		if s[len(s)-1] == '%' {
			i, err := strconv.Atoi(s[:len(s)-1])
			if err != nil {
				config.Log.Warn().Err(err).
					Str("actor", actor.Name).
					Str("config-entry", ConfigEntryDelayIncreaseFactor).
					Int("default-value", ConfigEntryDelayIncreaseFactorDefaultValue).
					Msg("Could not parse percentage. The default value will be used")
			} else {
				if i > 0 {
					delayIncreaseFactor = i
				} else {
					config.Log.Warn().Err(err).
						Str("actor", actor.Name).
						Str("config-entry", ConfigEntryDelayIncreaseFactor).
						Int("default-value", ConfigEntryDelayIncreaseFactorDefaultValue).
						Msg("Increase factor must be positive. The default value will be used")
				}
			}
		} else {
			f, err := strconv.ParseFloat(s, 64)
			if err != nil {
				config.Log.Warn().Err(err).
					Str("actor", actor.Name).
					Str("config-entry", ConfigEntryDelayIncreaseFactor).
					Int("default-value", ConfigEntryDelayIncreaseFactorDefaultValue).
					Msg("Could not parse float value. The default value will be used")
			} else {
				if f > 0 {
					delayIncreaseFactor = int(math.Ceil(f * 100))
				} else {
					config.Log.Warn().Err(err).
						Str("actor", actor.Name).
						Str("config-entry", ConfigEntryDelayIncreaseFactor).
						Int("default-value", ConfigEntryDelayIncreaseFactorDefaultValue).
						Msg("Increase factor must be positive. The default value will be used")
				}
			}
		}
	}

	return &ActorHandler{
		actor: actor,

		DelayMax:            delayMax,
		DelayIncreaseFactor: delayIncreaseFactor,

		log:  config.Log,
		conn: config.Conn,

		actorStorage:    config.ActorStorage,
		processManager:  config.ProcessManager,
		envelopeStorage: config.EnvelopeStorage,

		envWatcher: config.EnvelopeWatcher,
		director:   config.Director,

		actorMap: config.ActorMap,

		queue: newJobQueue(config.JobStorage, actor.ID, 1024),
		ctx:   ctx,

		running: make(map[api.UUID][]*Job),
	}
}

// ActorHandler is responsible for calling the 'Process' method of the actors,
// and handle the results
type ActorHandler struct {
	actor api.Actor

	DelayMax            time.Duration
	DelayIncreaseFactor int // expressed in 0.01

	log  zerolog.Logger
	conn *nats.Conn

	actorStorage    storage.ActorStorage
	processManager  *process.Manager
	envelopeStorage storage.EnvelopeStorage

	envWatcher *EnvelopeWatcher

	director directorServer

	queue           *JobQueue
	ctx             context.Context
	workerCtx       context.Context
	workerCtxCancel func()

	actorMap *svc_tools.ActorMap

	running      map[api.UUID][]*Job
	runningMutex sync.Mutex

	started bool

	wg sync.WaitGroup
}

func (h *ActorHandler) Start() {
	h.started = true

	h.workerCtx, h.workerCtxCancel = context.WithCancel(context.Background())

	// add a worker on the queue
	// TODO for parallel-able actors, start more routine
	h.wg.Add(1)

	go h.worker()
}

func (h *ActorHandler) Stop() {
	h.started = false

	h.workerCtxCancel()
	h.wg.Wait()
}

func (h *ActorHandler) Close() {
	if h.started {
		h.Stop()
	}
}

func (h *ActorHandler) worker() {
	defer h.wg.Done()

	for {
		select {
		case job := <-h.queue.C():
			if !h.handleJob(job) {
				h.queue.PushBack(job)
			}
		case <-h.workerCtx.Done():
			return
		}
	}
}

func (h *ActorHandler) Submit(job *Job) {
	if job.Node.ActorID != h.actor.ID {
		panic(fmt.Sprintf(
			"ActorHandler: Submit was given a job for a foreign actor. Expected %s (%s), got %s",
			h.actor.ID, h.actor.Name, job.Node.ActorID,
		))
	}
	h.queue.Add(job)
}

// handleJob handles a single job, and returns true if the job is handled (or
// ignored). If it returns 'false', then the job must be rescheduled
func (h *ActorHandler) handleJob(job *Job) bool {
	ctx, softCtx := h.setJobRunning(job)
	defer h.unsetJobRunning(job)

	jh := JobRunner{
		ActorHandler: h,
		job:          job,
		ctx:          ctx,
		softCtx:      softCtx,
		log: h.log.With().
			Str("job-id", job.JobID.String()).
			Logger(),
	}
	return jh.run()
}

func (h *ActorHandler) setJobRunning(job *Job) (context.Context, context.Context) {
	h.runningMutex.Lock()
	defer h.runningMutex.Unlock()

	ctx, cancel := context.WithCancel(h.workerCtx)
	job.cancel = cancel

	softCtx, softCancel := context.WithCancel(ctx)
	job.softCancel = softCancel

	processJobList := h.running[job.ProcessID]
	h.running[job.ProcessID] = append(processJobList, job)

	return ctx, softCtx
}

func (h *ActorHandler) unsetJobRunning(job *Job) {
	h.runningMutex.Lock()
	defer h.runningMutex.Unlock()

	job.cancel()

	processJobList := h.running[job.ProcessID]
	for i, j := range processJobList {
		if j == job {
			processJobList = append(
				processJobList[:i], processJobList[i+1:]...)
			break
		}
	}
	if len(processJobList) == 0 {
		delete(h.running, job.ProcessID)
	} else {
		h.running[job.ProcessID] = processJobList
	}
}

func (h *ActorHandler) CancelProcessJobs(id api.UUID, soft bool) {
	h.runningMutex.Lock()
	defer h.runningMutex.Unlock()

	processJobList := h.running[id]
	for _, job := range processJobList {
		if soft {
			job.softCancel()
		} else {
			job.cancel()
		}
	}
}
