package director

import (
	"context"
	"errors"
	"fmt"
	"strings"
	"time"

	"github.com/nats-io/nats.go"
	nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"github.com/rs/zerolog/log"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/tools"

	"xbus.io/xbus/pipeline"
	"xbus.io/xbus/storage"
)

type JobRunner struct {
	*ActorHandler
	job *Job

	pipeline *pipeline.Pipeline

	processStart time.Time
	processEnd   time.Time

	ctx     context.Context
	softCtx context.Context

	log zerolog.Logger
}

func (r *JobRunner) getAgent() *api.ActorAgentClient {
	nc := tools.NewServerToClientConn(r.conn, r.actor.AccountID.String())
	return api.NewActorAgentClient(nc, r.actor.ID.String())
}

func (r *JobRunner) initEnvelopeWatcher() *Subscription {
	return r.envWatcher.Watch(func(event api.EnvelopeEvent) {
		switch event.Status {
		case api.EnvelopeEvent_STALLED, api.EnvelopeEvent_ERROR:
			// Cancel any current call
			r.job.cancel()

			// Cancel the whole process
			process, err := r.processManager.Get(r.job.JobID.ProcessID)
			if err != nil {
				r.log.Err(err).
					Str("process-id", r.job.JobID.ProcessID.String()).
					Msg("error getting process")
			}

			if err := process.Cancel(
				fmt.Sprintf("Envelope %s status is %s", event.EnvelopeID, event.Status),
			); err != nil {
				r.log.Err(err).
					Str("process-id", r.job.JobID.ProcessID.String()).
					Msg("error canceling process")
			}

		// TODO Mark all the unfinished envelopes as error so they cannot receive
		// more fragments (not here actually, but in the process cancellation handling)
		default:
		}
	}, r.job.EnvelopeIDs...)
}

func (r *JobRunner) getPipeline() *pipeline.Pipeline {
	if r.pipeline != nil {
		return r.pipeline
	}
	// XXX The following lines loads the process and pipeline only to get
	// the pipeline name and version for metrics. It may slow down things.
	process, ok := r.processManager.ProcessStorage.GetProcess(r.job.ProcessID)
	if !ok {
		r.log.Error().
			Str("process-id", r.job.ProcessID.String()).
			Msg("unknown process id. job ignored")
		return nil
	}

	pipeline, ok := r.processManager.PipelineManager.GetPipeline(process.PipelineID)
	if !ok {
		r.log.Error().
			Str("pipeline-id", process.PipelineID.String()).
			Msg("unknown pipeline id. canceling job")
		r.processError("unknown pipeline: %s", process.PipelineID.String())
		return nil
	}
	r.pipeline = pipeline
	return r.pipeline
}

func (r *JobRunner) processIsRunning() bool {
	running := r.processManager.ProcessStorage.GetState(r.job.ProcessID).Status <= api.Process_RUNNING
	if !running {
		r.log.Debug().
			Str("process-id", r.job.ProcessID.String()).
			Msg("process is not running. job ignored")
	}
	return running
}

func (r *JobRunner) actorIsOnline() bool {
	online, _ := r.actorMap.IsOnline(r.actor.ID)
	return online
}

// loadInputEnvelopes load the input envelope first fragments of a job
func (r *JobRunner) loadInputEnvelopes() ([]api.UUID, []*api.ActorProcessRequest_Input, error) {
	var (
		inputs = r.processManager.ProcessStorage.GetInputs(
			r.job.ProcessID, r.job.Node)
		envelopeIDs   = make([]api.UUID, len(inputs))
		requestInputs = make([]*api.ActorProcessRequest_Input, len(inputs))
	)
	for i, input := range inputs {
		envelope, pos, err := r.envelopeStorage.ReadEnvelope(
			input.EnvelopeID, storage.EnvelopeStartPos(), 100*1024)
		if err != nil {
			log.Err(err).
				Str("envelope-id", input.EnvelopeID.String()).
				Msg("could not read envelope")
			return nil, nil, fmt.Errorf("could not read envelope %s: %w", input.EnvelopeID.String(), err)
		}
		requestInputs[i] = &api.ActorProcessRequest_Input{
			Name:     input.Name,
			Envelope: &envelope,
			Position: storage.EncodePosition(envelope.ID, pos),
		}

		envelopeIDs[i] = envelope.ID
	}
	return envelopeIDs, requestInputs, nil
}

func (r *JobRunner) makeRequest() *api.ActorProcessRequest {
	envelopeIDs, inputs, err := r.loadInputEnvelopes()
	if err != nil {
		r.processError(err.Error())
		return nil
	}
	r.job.EnvelopeIDs = envelopeIDs
	return &api.ActorProcessRequest{
		Context: &api.ProcessingContext{
			ProcessID: r.job.ProcessID,
			NodeID:    r.job.Node.NodeID,
		},
		Inputs: inputs,
	}
}

func (r *JobRunner) run() bool {
	r.log.Debug().Msg("handling job")

	if !r.actorIsOnline() {
		// job not handled and needs rescheduling
		return false
	}

	if !r.processIsRunning() {
		return true
	}

	if r.getPipeline() == nil {
		return true
	}

	request := r.makeRequest()
	if request == nil {
		return true
	}

	r.job.Status = api.ActorProcessingState_RUNNING

	r.job.SaveState()

	var handled bool

	defer func() {
		if !handled {
			r.job.Status = api.ActorProcessingState_PENDING
			r.job.SaveState()
		} else if r.job.Status == api.ActorProcessingState_RUNNING &&
			!r.job.Detached {
			r.job.Status = api.ActorProcessingState_ERROR
			r.job.SaveState()
		}
	}()

	handled = r.callProcessAttempts(request)

	return handled
}

func (r *JobRunner) callProcessAttempts(request *api.ActorProcessRequest) bool {
	// timeoutCount only apply to the current set of attempts
	r.job.timeoutCount = 0
	r.job.AttemptCount = 0

	watchSub := r.initEnvelopeWatcher()
	defer watchSub.Unsubscribe()

	for {
		err := r.callProcess(request)
		if err == nil {
			r.handleSuccess()
			return true
		}

		r.log.Info().Err(err).Str("actor", r.actor.Name).Msg("Process failed")
		var nrpcErr *nrpc.Error
		switch {
		case errors.Is(err, nats.ErrTimeout):
			if !r.handleTimeoutErr() {
				return false
			}
		case errors.As(err, &nrpcErr) && nrpcErr.GetType() == nrpc.Error_SERVERTOOBUSY:
			if r.handleServerTooBusy(nrpcErr) {
				return true
			}
		case errors.Is(err, nrpc.ErrCanceled):
			r.handleCanceled(err)
			return true
		default:
			r.handleUnexpectedError(err)
			return true
		}
	}
}

func (r *JobRunner) callProcess(request *api.ActorProcessRequest) error {
	r.log.Debug().
		Str("actor", r.actor.Name).
		Uint("attempt", r.job.AttemptCount).
		Msg("calling Process")

	r.processStart = time.Now()
	if r.job.FirstAttempt.IsZero() {
		r.job.FirstAttempt = r.processStart
		r.job.SaveState()
	}

	err := r.getAgent().Process(r.ctx, *request, r.handleProcessResponse)

	r.processEnd = time.Now()

	r.job.AttemptCount++

	r.job.SaveState()

	return err
}

func (r *JobRunner) handleSuccess() {
	status := strings.ToLower(r.job.Status.String())

	if r.job.Status == api.ActorProcessingState_RUNNING {
		// the job is detached, and will report its status asynchronously
		// with the Director.ProcessingEnd API entry point
		r.job.Detached = true
		r.job.SaveState()

		status = "detached"
	}
	r.observeJobDurations(status)

	if err := r.actorStorage.JustSeen(r.job.Node.ActorID); err != nil {
		r.log.Err(err).
			Str("actor", r.job.Node.ActorID.String()).
			Msg("error calling JustSeen")
	}
}

func (r *JobRunner) handleCanceled(err error) {
	r.processWarning("job canceled: %s", err)
	r.log.Debug().
		Err(err).
		Str("actor", r.actor.Name).
		Uint("attempt", r.job.AttemptCount).
		Msg("Error calling agent Process")
	r.observeJobDurations("call canceled")
}

func (r *JobRunner) handleProcessResponse(ctx context.Context, state api.ActorProcessingState) {
	r.job.Status = state.Status
	r.log.Debug().
		Interface("jobstate", r.job.JobState).
		Msg("saving job new state")
	if state.ActorLeaving {
		r.actorMap.SetOnline(r.job.Node.ActorID, false)
	}
	r.job.SaveState()
	switch state.Status {
	case api.ActorProcessingState_RUNNING:
	default:
		if err := r.director.ProcessingEnd(ctx, r.job.Node.ActorID.String(), state); err != nil {
			r.log.Err(err).Msg("error calling ProcessingEnd")
		}
	}
}

func (r *JobRunner) handleServerTooBusy(err *nrpc.Error) bool {
	// Back to pending, in case it is all restarted
	r.job.Status = api.ActorProcessingState_PENDING
	r.job.SaveState()
	r.job.timeoutCount = 0
	delay := time.Second
	for i := 0; i < int(r.job.AttemptCount)-1; i++ {
		delay += delay * time.Duration(r.DelayIncreaseFactor) / 100
		if delay >= r.DelayMax {
			delay = r.DelayMax
			break
		}
	}
	r.processWarning(
		"Actor too busy. Waiting %s before next attempt. Reason given: %s",
		delay, err.GetMessage())
	r.log.Info().
		Str("actor", r.actor.Name).
		Dur("duration", delay).
		Str("reason", err.Message).
		Msg("Agent cannot handle Process() now, wait some time before retrying.")
	select {
	case <-r.softCtx.Done():
		return true
	case <-time.After(delay):
		// Set to running because a new attempt will be made
		r.job.Status = api.ActorProcessingState_RUNNING
		r.job.SaveState()
	}
	return false
}

func (r *JobRunner) handleTimeoutErr() bool {
	r.job.timeoutCount++
	r.log.Error().
		Str("actor", r.actor.Name).
		Uint("count", r.job.timeoutCount).
		Msg("agent unresponsive")
	r.processWarning("agent is unresponsive, %d time", r.job.timeoutCount)
	if r.job.timeoutCount == 1 {
		r.log.Info().
			Str("actor", r.actor.Name).
			Msg("marking actor unresponsive")
		if err := r.actorStorage.MarkUnresponsive(r.job.Node.ActorID, r.job.FirstAttempt); err != nil {
			r.log.Err(err).Msg("failed to mark unresponsive")
		}
	} else if r.job.timeoutCount >= 5 {
		r.log.Error().
			Str("actor", r.actor.Name).
			Msg("marking actor offline after 5 timeouts on process attempts")
		if err := r.actorStorage.SetOnline(r.job.Node.ActorID, false); err != nil {
			r.log.Err(err).Msg("error marking offline")
		}
		// job not handled
		return false
	}
	r.log.Debug().
		Dur("duration", time.Duration(r.job.timeoutCount)*250).
		Msg("Sleeping")
	return true
}

func (r *JobRunner) handleUnexpectedError(err error) {
	r.processError("unexpected low-level error: %s", err)
	r.log.Err(err).
		Str("actor", r.actor.Name).
		Uint("attempt", r.job.AttemptCount).
		Msg("unexpected low-level error calling agent Process")
	r.observeJobDurations("unexpected error")
}

func (r *JobRunner) processError(msg string, args ...interface{}) {
	if err := r.director.ProcessingEnd(r.ctx, r.actor.ID.String(), api.ActorProcessingState{
		Context: api.ProcessingContext{ProcessID: r.job.ProcessID, NodeID: r.job.Node.NodeID},
		Status:  api.ActorProcessingState_ERROR,
		Messages: []api.LogMessage{{
			Time:  time.Now().UTC(),
			Level: api.LogLevel_ERROR,
			Text:  fmt.Sprintf(msg, args...),
		}},
	}); err != nil {
		log.Err(err).Msg("error calling ProcessingEnd")
	}
}

func (r *JobRunner) processWarning(msg string, args ...interface{}) {
	text := fmt.Sprintf(msg, args...)
	if err := r.director.Log(r.ctx, r.actor.ID.String(), api.ActorLogRequest{
		Context: &api.ProcessingContext{
			ProcessID: r.job.ProcessID,
			NodeID:    r.job.Node.NodeID,
		},
		EnvelopeID: &r.job.EnvelopeIDs[0],
		Messages: []api.LogMessage{{
			Time:  time.Now().UTC(),
			Level: api.LogLevel_WARNING,
			Text:  text,
		}},
	}); err != nil {
		r.log.Err(err).
			Str("actor", r.actor.Name).
			Str("process_id", r.job.ProcessID.String()).
			Str("text", text).
			Msg("Failed to log a warning message in the process")
	}
}

func (r *JobRunner) observeJobDurations(status string) {
	firstattempt := r.job.FirstAttempt
	jobCreatedAt := r.job.CreatedAt

	waitDuration := firstattempt.Sub(jobCreatedAt)
	busyDuration := r.processStart.Sub(firstattempt)
	processDuration := r.processEnd.Sub(r.processStart)
	totalDuration := waitDuration + busyDuration + processDuration

	labels := prometheus.Labels{
		"actor":            r.actor.Name,
		"pipeline":         r.pipeline.Name,
		"pipeline_version": r.pipeline.Version,
		"node":             r.job.Node.NodeID,
		"status":           status,
	}

	jobWaitTime.With(labels).Observe(waitDuration.Seconds())
	jobBusyTime.With(labels).Observe(busyDuration.Seconds())
	jobProcessTime.With(labels).Observe(processDuration.Seconds())
	jobTotalTime.With(labels).Observe(totalDuration.Seconds())
}

var (
	jobWaitTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_wait_time_seconds",
			Help:       "The actor job wait in queue before effective process, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)

	jobBusyTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_busy_time_seconds",
			Help:       "The actor busy time before effective process, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)

	jobProcessTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_process_time_seconds",
			Help:       "The actor job process time, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)

	jobTotalTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_total_time_seconds",
			Help:       "The actor job wait and execution time, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)
)

func init() {
	prometheus.MustRegister(jobWaitTime)
	prometheus.MustRegister(jobBusyTime)
	prometheus.MustRegister(jobProcessTime)
	prometheus.MustRegister(jobTotalTime)
}
