package metrics

import (
	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v3/api/control"
)

// NewLogStatCollector creates a new LogStatCollector
func NewLogStatCollector(
	control *control.Client, log zerolog.Logger,
) *LogStatCollector {
	c := LogStatCollector{
		control: control,
		log:     log,

		storageCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "storage", "count"),
			"Number of items in the storage",
			[]string{"name", "implementation"},
			prometheus.Labels{"type": "log"},
		),
		storageSizeDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "storage", "size"),
			"Total size of the storage",
			[]string{"name", "implementation"},
			prometheus.Labels{"type": "log"},
		),
	}
	return &c
}

// LogStatCollector implements the prometheus.Collector interface
type LogStatCollector struct {
	control *control.Client
	log     zerolog.Logger

	storageCountDesc *prometheus.Desc
	storageSizeDesc  *prometheus.Desc
}

// Describe implements the prometheus.Collector interface.
func (c LogStatCollector) Describe(ch chan<- *prometheus.Desc) {
	ch <- c.storageCountDesc
	ch <- c.storageSizeDesc
}

// Collect implements the prometheus.Collector interface.
func (c *LogStatCollector) Collect(ch chan<- prometheus.Metric) {
	stat, err := c.control.Logs.GetStorageStat()
	if err != nil {
		c.log.Warn().Err(err).Msg("metrics: error reading log storage stats")
		return
	}

	for _, entry := range stat.Entries {
		ch <- prometheus.MustNewConstMetric(
			c.storageCountDesc,
			prometheus.GaugeValue,
			float64(entry.Count),
			entry.Name,
			entry.Implementation,
		)

		ch <- prometheus.MustNewConstMetric(
			c.storageSizeDesc,
			prometheus.GaugeValue,
			float64(entry.EstimatedSize),
			entry.Name,
			entry.Implementation,
		)
	}
}
