// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"encoding/json"
	"errors"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"github.com/slicebit/qb"
	"xbus.io/go-xbus/v3/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql/model"
)

// NewActorStorage initialize a new ActorStorage
func NewActorStorage(model *model.Model, db *yago.DB, logger zerolog.Logger) *ActorStorage {
	return &ActorStorage{
		Base: Base{
			model:  model,
			db:     db,
			logger: logger,
		},
		onActorUp:   storage.NoopActorEventHandler,
		onActorDown: storage.NoopActorEventHandler,
	}
}

// ActorStorage implements all the persistor interfaces
type ActorStorage struct {
	Base

	onActorUp   storage.ActorEventHandler
	onActorDown storage.ActorEventHandler
}

func actorModelToAPI(actor model.Actor) api.Actor {
	result := api.Actor{
		ID:           api.UUID(actor.ID),
		Name:         actor.Name,
		AccountID:    api.UUID(actor.AccountID.UUID),
		Kind:         actor.Kind,
		Status:       actor.Status,
		Roles:        []string{actor.Role},
		Online:       actor.Online,
		Unresponsive: actor.Unresponsive,
		LastSeen:     actor.LastSeen,
	}
	if actor.Config != "" {
		if err := json.Unmarshal([]byte(actor.Config), &result.Config); err != nil {
			panic(err)
		}
	}
	return result
}

func actorAPIToModel(actor api.Actor) model.Actor {
	config, err := json.Marshal(actor.Config)
	if err != nil {
		panic(err)
	}
	return model.Actor{
		Base:   model.Base{ID: uuid.UUID(actor.ID)},
		Name:   actor.Name,
		Kind:   actor.Kind,
		Role:   actor.Roles[0],
		Status: actor.Status,
		AccountID: uuid.NullUUID{
			Valid: !actor.AccountID.IsZero(), UUID: uuid.UUID(actor.AccountID)},
		Config: string(config),
	}
}

func actorListModelToAPI(actors []model.Actor) (out []api.Actor) {
	for _, actor := range actors {
		out = append(out, actorModelToAPI(actor))
	}
	return
}

// SetEventHandlers sets the event handlers
func (p *ActorStorage) SetEventHandlers(onActorUp, onActorDown storage.ActorEventHandler) {
	if onActorUp == nil {
		p.onActorUp = storage.NoopActorEventHandler
	} else {
		p.onActorUp = onActorUp
	}
	if onActorDown == nil {
		p.onActorDown = storage.NoopActorEventHandler
	} else {
		p.onActorDown = onActorDown
	}
}

// Get returns an actor from its id
func (p ActorStorage) Get(id api.UUID) (api.Actor, bool) {
	var (
		actor model.Actor
	)
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	err := tx.Query(p.model.Actor).Get(&actor, uuid.UUID(id))
	if errors.Is(err, yago.ErrRecordNotFound) {
		return api.Actor{}, false
	}
	if err != nil {
		panic(err)
	}
	return actorModelToAPI(actor), true
}

// GetByName returns an actor from its name
func (p ActorStorage) GetByName(name string) (api.Actor, bool) {
	var (
		actor model.Actor
	)
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()
	err := tx.Query(p.model.Actor).Where(p.model.Actor.Name.Eq(name)).One(&actor)
	if errors.Is(err, yago.ErrRecordNotFound) {
		return api.Actor{}, false
	}
	if err != nil {
		panic(err)
	}
	return actorModelToAPI(actor), true
}

func (p ActorStorage) list(clauses ...qb.Clause) []api.Actor {
	var actors []model.Actor

	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	q := tx.Query(p.model.Actor)
	if len(clauses) != 0 {
		q = q.Where(clauses...)
	}

	must(q.All(&actors))

	return actorListModelToAPI(actors)
}

// List returns all the actors
func (p ActorStorage) List() []api.Actor {
	return p.list()
}

// ListAccount returns a list of all actors the specified account provides access to.
func (p ActorStorage) ListAccount(accountID api.UUID) []api.Actor {
	return p.list(
		p.model.Actor.AccountID.Eq(uuid.UUID(accountID)),
	)
}

// ListKind returns a list of all actors of one type
func (p ActorStorage) ListKind(actorKind api.Actor_Kind) []api.Actor {
	return p.list(
		p.model.Actor.Kind.Eq(actorKind),
	)
}

// ListRole returns a list of all actors of one role
func (p ActorStorage) ListRole(actorKind api.Actor_Kind, role string) []api.Actor {
	return p.list(
		p.model.Actor.Kind.Eq(actorKind),
		p.model.Actor.Role.Eq(role),
	)
}

// Create creates an Actor
func (p ActorStorage) Create(actor api.Actor) (api.Actor, error) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	dbActor := actorAPIToModel(actor)
	if err := tx.Insert(&dbActor); err != nil {
		// TODO identify constraint errors and return them (not panic)
		panic(err)
	}
	tx.MustCommit()
	return actorModelToAPI(dbActor), nil
}

// Delete deletes an Actor
func (p ActorStorage) Delete(id api.UUID) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	must(
		tx.Delete(&model.Actor{Base: model.Base{ID: uuid.UUID(id)}}))
	tx.MustCommit()
}

// Update updates an Actor
func (p ActorStorage) Update(actor api.Actor) (api.Actor, error) {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	dbActor := actorAPIToModel(actor)
	if err := tx.Update(&dbActor); err != nil {
		// TODO identify constraint errors and return them
		panic(err)
	}
	tx.MustCommit()
	return actorModelToAPI(dbActor), nil
}

// JustSeen update the actor lastSeen timestamp
func (p ActorStorage) JustSeen(id api.UUID) error {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var actor model.Actor
	if err := tx.Query(p.model.Actor).Get(&actor, uuid.UUID(id)); errors.Is(err, yago.ErrRecordNotFound) {
		return err
	} else if err != nil {
		panic(err)
	}
	actor.Unresponsive = false
	actor.LastSeen = time.Now().UTC()
	if err := tx.Update(&actor, model.ActorUnresponsive, model.ActorLastSeen); err != nil {
		// TODO if constraint error, return a user friendly error
		panic(err)
	}
	tx.MustCommit()
	return nil
}

// MarkUnresponsive marks the actor as unresponsive if the actor
// was not seen since decisionTime
func (p ActorStorage) MarkUnresponsive(id api.UUID, decisionTime time.Time) error {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var actor model.Actor
	err := tx.Query(p.model.Actor).Get(&actor, uuid.UUID(id))
	if errors.Is(err, yago.ErrRecordNotFound) {
		return err
	}
	if err != nil {
		panic(err)
	}
	if actor.LastSeen.After(decisionTime) {
		// The actor was seen after this decision was made, so we ignore it
		p.logger.Info().
			Str("actor", actor.Name).
			Time("lastseen", actor.LastSeen).
			Time("decision-time", decisionTime).
			Msg("MarkUnresponsive: Actor was last seen after decision time")
		return nil
	}
	if !actor.Unresponsive {
		actor.Unresponsive = true
		err = tx.Update(&actor, model.ActorUnresponsive)
		if err != nil {
			panic(err)
		}
	}
	tx.MustCommit()
	return nil
}

// SetOnline change the online status of the actor
func (p ActorStorage) SetOnline(id api.UUID, online bool) error {
	tx := p.MustBegin()
	defer tx.RollbackIfOpened()

	var actor model.Actor
	err := tx.Query(p.model.Actor).Get(&actor, uuid.UUID(id))
	if errors.Is(err, yago.ErrRecordNotFound) {
		return storage.ErrNoSuchID
	}
	if err != nil {
		panic(err)
	}

	changed := actor.Online != online
	if changed || online && actor.Unresponsive {
		actor.Online = online
		if online {
			actor.Unresponsive = false
		}
		err = tx.Update(&actor, model.ActorOnline, model.ActorUnresponsive)
		if err != nil {
			panic(err)
		}
		tx.MustCommit()
		if changed {
			if online {
				p.logger.Info().
					Str("actor", actor.Name).
					Msg("set actor UP")
				if err := p.JustSeen(id); err != nil {
					p.logger.Err(err).Msg("Error calling JustSeen")
				}
				p.onActorUp(actorModelToAPI(actor))
			} else {
				p.logger.Info().
					Str("actor", actor.Name).
					Msg("set actor DOWN")
				p.onActorDown(actorModelToAPI(actor))
			}
		}
	}
	return nil
}
