package sql

import (
	"database/sql"
	"errors"
	"fmt"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/storage/sql/model"
)

// SessionStorage implements storage.SessionStorage
type SessionStorage struct {
	Base
}

// NewSessionStorage implements session storage
func NewSessionStorage(db *yago.DB, logger zerolog.Logger) *SessionStorage {
	return &SessionStorage{
		Base: Base{
			model:  nil,
			db:     db,
			logger: logger,
		},
	}
}

// List returns all the opened sessions
func (store *SessionStorage) List() ([]storage.Session, error) {
	tx := store.MustBegin()
	defer tx.BlindRollback()
	rows, err := tx.GetEngine().Query(model.RawSQL(
		"SELECT token, account_id, valid_until FROM account_session",
	))
	if err != nil {
		return nil, err
	}
	var sessionList []storage.Session
	if err := func() (err error) {
		defer func() {
			if e := rows.Close(); e != nil && err == nil {
				err = e
			}
		}()

		for rows.Next() {
			var session storage.Session
			if err := rows.Scan(
				&session.Token,
				(*uuid.UUID)(&session.AccountID),
				&session.ValidUntil,
			); err != nil {
				return err
			}
			session.ValidUntil = session.ValidUntil.UTC()
			sessionList = append(sessionList, session)
		}
		return nil
	}(); err != nil {
		return nil, err
	}
	return sessionList, nil
}

// Get returns a Session from its token
func (store *SessionStorage) Get(token string) (*storage.Session, error) {
	tx := store.MustBegin()
	defer tx.BlindRollback()
	row := tx.GetEngine().QueryRow(model.RawSQL(
		"SELECT token, account_id, valid_until "+
			"FROM account_session "+
			"WHERE token = $1",
		token,
	))
	var session storage.Session
	if err := row.Scan(
		&session.Token,
		(*uuid.UUID)(&session.AccountID),
		&session.ValidUntil,
	); err != nil {
		if errors.Is(err, sql.ErrNoRows) {
			return nil, fmt.Errorf("%w: session '%s'", storage.ErrNotExist, token)
		}
		return nil, err
	}
	session.ValidUntil = session.ValidUntil.UTC()
	return &session, nil
}

// Set create or update a Session
func (store *SessionStorage) Set(session storage.Session) error {
	tx := store.MustBegin()
	defer tx.RollbackIfOpened()

	row := tx.GetEngine().QueryRow(model.RawSQL(
		"INSERT INTO account_session (token, account_id, valid_until) "+
			"VALUES ($1, $2, $3) "+
			"ON CONFLICT (token) "+
			"DO UPDATE SET valid_until = $3 "+
			"RETURNING (account_id)",
		session.Token,
		uuid.UUID(session.AccountID),
		session.ValidUntil,
	))

	var accountID uuid.UUID

	if err := row.Scan(&accountID); err != nil {
		return err
	}

	if accountID != uuid.UUID(session.AccountID) {
		return fmt.Errorf(
			"Token %s already exists and is not associated to the account %s, "+
				"but to the account %s",
			session.Token, session.AccountID, accountID,
		)
	}

	return tx.Commit()
}

// Delete deletes a token
func (store *SessionStorage) Delete(token string) error {
	tx := store.MustBegin()
	defer tx.RollbackIfOpened()

	_, err := tx.GetEngine().Exec(model.RawSQL(
		"DELETE FROM account_session WHERE token = $1",
		token,
	))
	if err != nil {
		return err
	}

	return tx.Commit()
}

// DeleteOlderThan deletes the sessions based on their 'validUntil'
func (store *SessionStorage) DeleteOlderThan(time time.Time) error {
	tx := store.MustBegin()
	defer tx.RollbackIfOpened()

	_, err := tx.GetEngine().Exec(model.RawSQL(
		"DELETE FROM account_session "+
			"WHERE valid_until < $1",
		time,
	))
	if err != nil {
		return err
	}

	return tx.Commit()
}
